@@ -1,4 +1,4 @@
-(* $Id: frontend.ml 116 2007-11-11 22:40:21Z gerd $
+(* $Id: frontend.ml 130 2009-02-16 01:09:27Z gerd $
  * ----------------------------------------------------------------------
  *
  *)
@@ -21,6 +21,47 @@ type psubst =
 ;;
 
 
+
+let slashify s =
+  match Findlib_config.system with
+    | "mingw" | "cygwin" ->
+	let u = String.copy s in
+	for k = 0 to String.length u - 1 do
+	  if u.[k] = '\\' then u.[k] <- '/'
+	done;
+	u
+    | _ ->
+	s
+
+
+let out_path ?(prefix="") s =
+  match Findlib_config.system with
+    | "mingw" | "cygwin" ->
+	let u = slashify s in
+	prefix ^ 
+	  (if String.contains u ' ' then
+	     (* Desperate attempt to fix the space problem in paths.
+                Note that we invoke commands via Unix.open_process, and
+                this function already quotes the arguments on win32.
+                However, for -ccopt arguments, one quoting level seems
+                to be lost, and we have to add another level to compensate.
+                E.g. for the list of args
+                  [ -ccopt; -L/my programs/include -L/somewhere ]
+                we get after out_path
+                  [ -ccopt; "-I/my programs/include -L/somewhere" ]
+                which actually translates to
+                  -ccopt "\"-I/my programs/include\" \"-L/somewhere\""
+                on the command line, i.e. a double-quoted argument.
+	      *)
+	     "\"" ^ u ^ "\""
+	   else
+	     u
+	  )
+    | _ ->
+	prefix ^ slashify s
+
+
+
 let percent_subst spec lookup s =
   (* spec = [ "%c", [ "ctext1"; "ctext2"; ... ];
    *          "%d", [ "dtext1"; "dtext2"; ... ] ]
@@ -118,7 +159,7 @@ let use_package prefix pkgnames =
   let pdirs =
     List.map
       (fun pname ->
-         "-I " ^ package_directory pname
+         "-I " ^ out_path(package_directory pname)
       )
       pkgnames
   in
@@ -354,7 +395,7 @@ let expand predicates eff_packages format =
 	    (* May raise No_such_package *)
 	 let spec =
 	   [ "%p",  [pkg];
-             "%d",  [dir];
+             "%d",  [out_path dir];
 	     "%D",  [try package_property predicates pkg "description"
 		     with Not_found -> "[n/a]"];
 	     "%v",  [try package_property predicates pkg "version"
@@ -561,7 +602,7 @@ let process_pp_spec syntax_preds packages pp_opts =
       (List.map
 	 (fun pkg ->
 	    let pkgdir = package_directory pkg in
-	      [ "-I"; pkgdir ]
+	      [ "-I"; slashify pkgdir ]
 	 )
 	 pp_packages) in
 
@@ -723,13 +764,13 @@ let ocamlc which () =
 		   (fun s ->
 		      let s = resolve_path s in
                       if Sys.file_exists s then incpath := s :: !incpath;  (* reverted below *)
-		      add_spec_fn "-I" s ));
+		      add_spec_fn "-I" (slashify s) ));
 
 	  "-impl", 
-	  Arg.String (fun s -> pass_files := !pass_files @ [ Impl s ]);
+	  Arg.String (fun s -> pass_files := !pass_files @ [ Impl(slashify s) ]);
 
 	  "-intf", 
-	  Arg.String (fun s -> pass_files := !pass_files @ [ Intf s ]);
+	  Arg.String (fun s -> pass_files := !pass_files @ [ Intf(slashify s) ]);
 
 	  "-pp", 
 	  Arg.String (fun s -> pp_specified := true; add_spec_fn "-pp" s);
@@ -916,8 +957,8 @@ let ocamlc which () =
 	    if List.mem npkgdir exclude_list then
 	      []
 	    else
-	      [ "-I"; pkgdir;
-		"-ccopt"; "-I" ^ pkgdir; ])
+	      [ "-I"; slashify pkgdir;
+		"-ccopt"; out_path ~prefix:"-I" pkgdir; ])
 	 eff_packages_dl) in
 
   let l_options =
@@ -930,9 +971,9 @@ let ocamlc which () =
 	    else
 	      if Findlib_config.system = "win32" then
 		(* Microsoft toolchain *)
-		[ "-ccopt"; "/link /libpath:" ^ pkgdir ]
+		[ "-ccopt"; out_path ~prefix:"/link /libpath:" pkgdir ]
 	      else
-		[ "-ccopt"; "-L" ^ pkgdir; ])
+		[ "-ccopt"; out_path ~prefix:"-L" pkgdir; ])
 	 eff_link_dl) in
 
   let archives =
@@ -949,6 +990,7 @@ let ocamlc which () =
 		 | `POSIX_threads -> threads_dir
 	     else
 	       package_directory pkg in
+	   let pkg_dir = slashify pkg_dir in
 	   List.map
 	     (fun arch ->
 		resolve_path ~base:pkg_dir arch)
@@ -999,13 +1041,13 @@ let ocamlc which () =
 
   let dll_dirs =
     remove_dups
-      ((List.map package_directory !dll_pkgs) @   (* XXX *)
+      ((List.map package_directory !dll_pkgs) @ 
        (if !dll_pkgs_all then eff_link_dl else [])) in
 
   let dll_options =
     List.flatten
       (List.map
-	 (fun pkg -> ["-dllpath";  pkg] )
+	 (fun pkg -> ["-dllpath";  slashify pkg] )
 	 dll_dirs) in
 
   let arguments =
@@ -1126,7 +1168,7 @@ let ocamldoc() =
     remove_dups (List.map package_directory eff_packages) in
 
   let arguments =
-    (List.flatten (List.map (fun d -> [ "-I"; d ]) eff_packages_dl)) @
+    (List.flatten (List.map (fun d -> [ "-I"; slashify d ]) eff_packages_dl)) @
     pp_command @
     !options in
 
@@ -1233,7 +1275,7 @@ let ocamldep () =
 	    add_switch
 	    add_spec
 	    [ "-I",
-	      Arg.String (fun s -> add_spec_fn "-I" (resolve_path s));
+	      Arg.String (fun s -> add_spec_fn "-I" (slashify (resolve_path s)));
 
 	      "-pp", Arg.String (fun s -> pp_specified := true;
 		 	           add_spec_fn "-pp" s);
@@ -1312,7 +1354,7 @@ let ocamlbrowser () =
 
   Arg.parse
       [
-	"-I", Arg.String (fun s -> add_spec_fn "-I" (resolve_path s)),
+	"-I", Arg.String (fun s -> add_spec_fn "-I" (slashify(resolve_path s))),
            " <dir>          Add <dir> to the list of include directories";
 	"-all", Arg.Set add_all,
 	     "              Add all packages to include path";
@@ -1333,7 +1375,7 @@ let ocamlbrowser () =
        (List.map
 	  (fun pkg ->
 	     let dir = Findlib.package_directory pkg in
-	     [ "-I"; dir ]
+	     [ "-I"; slashify dir ]
 	  )
 	  !packages
        )
@@ -1744,18 +1786,28 @@ let install_package () =
 
   if dll_list <> [] && have_libexec && !ldconf <> "ignore" then begin
     (* Check whether libexec is mentioned in ldconf *)
+    (* FIXME: We have to be careful with case-insensitive filesystems. 
+       Currently, we only check for Win32, but also OS X may have ci 
+       filesystems. So some better check would be nice.
+       Furthermore, String.lowercase assumes that the encoding of file names is
+       ISO-8859-1. This is probably plainly wrong.
+     *)
     let lines = read_ldconf !ldconf in
     let dlldir_norm = Fl_split.norm_dir dlldir in
-    if not (List.exists
-	      (fun d -> Fl_split.norm_dir d = dlldir_norm)
-	      lines) then
+    let dlldir_norm_lc = String.lowercase dlldir_norm in
+    let ci_filesys = (Sys.os_type = "Win32") in
+    let check_dir d =
+      let d' = Fl_split.norm_dir d in
+      (d' = dlldir_norm) || 
+	(ci_filesys && String.lowercase d' = dlldir_norm_lc) in
+    if not (List.exists check_dir lines) then
       prerr_endline("ocamlfind: [WARNING] You have installed DLLs but the directory " ^ dlldir_norm ^ " is not mentioned in ld.conf");
   end;
 
   (* Check if there is a postinstall script: *)
   let postinstall = Filename.concat !destdir "postinstall" in
   if Sys.file_exists postinstall then
-    run_command true postinstall [ !destdir; !pkgname ]
+    run_command true postinstall [ slashify !destdir; !pkgname ]
 ;;
 
 
@@ -1870,7 +1922,7 @@ let remove_package () =
   (* Check if there is a postremove script: *)
   let postremove = Filename.concat !destdir "postremove" in
   if Sys.file_exists postremove then
-    run_command true postremove [ !destdir; !pkgname ]
+    run_command true postremove [ slashify !destdir; !pkgname ]
 ;;
 
 
